using Microsoft.VisualBasic.CompilerServices;
using Raytracer;
using System;
using System.DrawingCore;
using System.Numerics;

namespace Raytracer
{
    public struct Vector3
    {
        public double x, y, z;
        public Vector3(double x, double y, double z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }
        public Vector3(double v)
        {
            this.x = v;
            this.y = v;
            this.z = v;
        }
        public Vector3 normalized
        {
            get
            {
                double l = System.Math.Sqrt(x * x + y * y + z * z);
                if (l == 0) return this;
                return this / l;
            }
        }
        public double sqrMagnitude
        {
            get
            {
                return x * x + y * y + z * z;
            }
        }
        public double magenitude
        {
            get
            {
                return Math.Sqrt(sqrMagnitude);
            }
        }
        public static Vector3 operator +(Vector3 a, Vector3 b)
        {
            return new Vector3(a.x + b.x, a.y + b.y, a.z + b.z);
        }
        public static Vector3 operator /(Vector3 a, double b)
        {
            return new Vector3(a.x / b, a.y / b, a.z / b);
        }
        public static Vector3 operator -(Vector3 a, Vector3 b)
        {
            return new Vector3(a.x - b.x, a.y - b.y, a.z - b.z);
        }
        public static Vector3 operator -(Vector3 a)
        {
            return new Vector3(-a.x , -a.y , -a.z);
        }
        //DOT
        public static Vector3 operator *(Vector3 a, double b)
        {
            return new Vector3(a.x * b, a.y * b, a.z * b);
        }
        public static Vector3 operator *(double b, Vector3 a)
        {
            return new Vector3(a.x * b, a.y * b, a.z * b);
        }
        public static Vector3 operator *(Vector3 a, Vector3 b)
        {
            return new Vector3(a.x * b.x, a.y * b.y, a.z * b.z);
        }
        public static Vector3 operator *(Vector3 v, Matrix4x4 m)
        {
            return new Vector3(
                v.x * m.M11 + v.y * m.M21 + v.z * m.M31 + m.M41,
                v.x * m.M12 + v.y * m.M22 + v.z * m.M32 + m.M42,
                v.x * m.M13 + v.y * m.M23 + v.z * m.M33 + m.M43);
        }

        public double Dot(Vector3 v)
        {
            return x * v.x + y * v.y + z * v.z;
        }
        public Vector3 Cross(Vector3 v)
        {
            return new Vector3(y * v.z - v.y * z, -x * v.z + v.x * z, x * v.y - v.x * y);
        }

        public static implicit operator Color(Vector3 v)
        {
            Color c = Color.FromArgb((int)(v.x * 255),(int)(v.y * 255),(int)(v.z * 255));
            return c;
        }
        public static implicit operator System.Numerics.Vector3(Vector3 v)
        {
            System.Numerics.Vector3  a = new System.Numerics.Vector3((float)v.x, (float)v.y, (float)v.z);
            return a;
        }
        public Vector3 vertical
        {
            get 
            {
                Vector3 v = new Vector3();
                if (z != 0)
                {
                    v.x = 1;
                    v.y = 1;
                    v.z = (-x - y) / z;
                }
                if (y != 0)
                {
                    v.x = 1;
                    v.z = 1;
                    v.y = (-x - z) / y;
                }
                if (x != 0)
                {
                    v.y = 1;
                    v.z = 1;
                    v.x = (-y - z) / x;
                }
                return v.normalized;
            }
        }
        public Vector3 RotateByAxis(Vector3 axis,double rad)
        {
            axis = axis.normalized;
            return this * Math.Cos(rad) + (axis.Cross(this) * Math.Sin(rad) + axis * (axis.Dot(this) * (1 - Math.Cos(rad))));
        }


        public override string ToString()
        {
            return x + "," + y + "," + z;
        }
    }

    public struct Vector2
    {
        public double x, y;
        public Vector2(double x, double y)
        {
            this.x = x;
            this.y = y;
        }
        public static implicit operator Vector2i(Vector2 v)
        {
            return new Vector2i((int)v.x, (int)v.y);
        }
    }
    public struct Vector4
    {
        public double x, y, z, w;
        public Vector4(double x, double y, double z, double w)
        {
            this.x = x;
            this.y = y;
            this.z = z;
            this.w = w;
        }
        public static implicit operator Vector4i(Vector4 i)
        {
            return new Vector4i((int)i.x, (int)i.y, (int)i.z, (int)i.w);
        }
        public static implicit operator Quaternion(Vector4 i)
        {
            return new Quaternion((float)i.x, (float)i.y, (float)i.z, (float)i.w);
        }
        public static implicit operator Vector4(Quaternion i)
        {
            return new Vector4(i.X, i.Y,i.Z,i.W);
        }
    }

    public struct Vector4i
    {
        public int x, y, z, w;
        public Vector4i(int x, int y, int z, int w)
        {
            this.x = x;
            this.y = y;
            this.z = z;
            this.w = w;
        }
        public static implicit operator Vector4(Vector4i i)
        {
            return new Vector4(i.x,i.y,i.z,i.w);
        }
    }

    public interface IHitable
    {
        public bool Hit(Ray ray);
    }

    public struct Bound:IHitable
    {
        public Vector3 min, max;
        public Vector3 center
        {
            get
            {
                return (min + max) / 2;
            }
        }
        public Vector3 size
        {
            get
            {
                return max - min;
            }
        }
        public Bound(Vector3 min, Vector3 max)
        {
            this.min = min;
            this.max = max;
        }

        public bool Hit(Ray ray)
        {
            return false;
        }
    }
    public struct Ray
    {
        public Vector3 origin, direction;
        public Ray(Vector3 o, Vector3 d)
        {
            origin = o;
            direction = d.normalized;
        }
    }
    public struct Vector2i
    {
        public int x, y;
        public Vector2i(int x, int y)
        {
            this.x = x;
            this.y = y;
        }
        public static implicit operator Vector2(Vector2i v) 
        {
            return new Vector2((double)v.x,(double)v.y);
        }
    }
    public struct Triangle
    {
        public Vector3 v1;
        public Vector3 v2;
        public Vector3 v3;
    }
}