using System;
using System.IO;


namespace Raytracer
{
    public struct HitInfo
    {
        public Vector3 position;
        public Vector3 normal;
        public Vector2 uv;
        public Vector3 color;
    }
    public class Surface
    {
        public ITexture fresnel;
        public ITexture alpha;
        public ITexture reflectionRate;
        public ITexture emission;
        public ITexture color;
        public bool isLight;
    }
    public abstract class Renderable : SceneObejct
    {
        public Bound bound;
        public Shader shader;
        public abstract bool HitTest(Ray ray,out HitInfo hitinfo);
    }
    public class Sphere : Renderable
    {
        double radius;
        public Sphere(double radius)
        {
            this.radius = radius;
            this.bound = new Bound(new Vector3(-radius), new Vector3(radius));
        }
        public Sphere(Vector3 position, double radius)
        {
            this.radius = radius;
            this.transform.position = position;
            this.bound = new Bound(new Vector3(-radius) + position, new Vector3(radius) + position);
        }
        public override bool HitTest(Ray ray,out HitInfo hitinfo)
        {
            hitinfo = new HitInfo();
            Vector3 op = transform.position - ray.origin;
            double l = op.Dot(ray.direction);
            double d = op.sqrMagnitude - l * l;
            double q = Math.Sqrt(radius * radius - d);
            double t = l - q,t0 = l + q;
            if (t < 0)
                t = t0;
            if (d <= radius * radius && t > 0.0001)
            {
                hitinfo = new HitInfo();
                hitinfo.position = ray.direction * t + ray.origin;
                hitinfo.normal = (hitinfo.position - transform.position).normalized;
                hitinfo.uv = new Vector2(0, 0);
                return true;
            }
            return false;
        }
    }

    public class InfinityPlane : Renderable
    {
        public Vector3 normal;
        public InfinityPlane(Vector3 pos,Vector3 normal)
        {
            this.normal = normal;
            transform.position = pos;
        }
        public override bool HitTest(Ray ray, out HitInfo hitinfo)
        {
            hitinfo = new HitInfo();
            double t = (normal.Dot(transform.position) - normal.Dot(ray.origin)) / (normal.Dot(ray.direction));
            if (t > 0.0001)
            {
                hitinfo.position = ray.origin + ray.direction * t;
                hitinfo.normal = normal;
                hitinfo.uv = new Vector2(0,0);
                return true;
            }
            return false;
        }
    }

    public class MeshRenderer : Renderable
    {
        public Mesh mesh;
        public MeshRenderer(Mesh mesh)
        {
            this.mesh = mesh;
        }
        public override bool HitTest(Ray ray, out HitInfo hitinfo)
        {
            hitinfo = new HitInfo();
            HitInfo nearest = hitinfo;
            Vector3 hitPos,normal;
            double nDistance = 99999999999999, dist;
            bool ret = false;
            for (int i = 0; i < mesh.indeces.Length; i += 3)
            {
                if (TriangleIntersection(ray,mesh.vertices[mesh.indeces[i]], mesh.vertices[mesh.indeces[i+1]], mesh.vertices[mesh.indeces[i+2]], out hitPos, out normal,out dist))
                {
                    if (dist < nDistance)
                    {
                        nDistance = dist;
                        nearest.normal = normal;
                        nearest.position = hitPos;
                        ret = true;
                    }
                }
            }
            hitinfo = nearest;
            return ret;
        }
        public static bool TriangleIntersection(Ray ray,Vector3 v0,Vector3 v1,Vector3 v2,out Vector3 position,out Vector3 normal,out double t)
        {
            t = 0;
            position = new Vector3(0);
            normal = position;
            double u, v;
            // E1
            Vector3 E1 = v1 - v0;

            // E2
            Vector3 E2 = v2 - v0;

            // P
            Vector3 P = ray.direction.Cross(E2);
            normal = E1.Cross(E2).normalized;

            // determinant
            double det = E1.Dot(P);

            // keep det > 0, modify T accordingly
            Vector3 T;
            if (det > 0)
            {
                T = ray.origin - v0;
            }
            else
            {
                T = v0 - ray.origin;
                det = -det;
            }

            // If determinant is near zero, ray lies in plane of triangle
            if (det < 0.0001f)
                return false;

            // Calculate u and make sure u <= 1
            u = T.Dot(P);
            if (u < 0.0f || u > det)
                return false;

            // Q
            Vector3 Q = T.Cross(E1);

            // Calculate v and make sure u + v <= 1
            v = ray.direction.Dot(Q);
            if (v < 0.0f || u + v > det)
                return false;

            // Calculate t, scale parameters, ray intersects triangle
            t = E2.Dot(Q);

            double fInvDet = 1.0 / det;
            t *= fInvDet;
            if (t < 0.0001) return false;
            u *= fInvDet;
            v *= fInvDet;

            position = ray.origin + t * ray.direction;
            return true;

        }
    }
    public class Mesh
    { 
        public Vector3[] vertices;
        public int[] indeces;
        public Vector2[] uv;
        public Mesh[] subMeshes;

        public void LoadFromFile(string path)
        {
            FileStream fs = new FileStream(path, FileMode.Open);
            BinaryReader sr = new BinaryReader(fs);
            int geoCount = sr.ReadInt32();
            GetMesh(sr, out indeces, out vertices, out uv);
            /*int[] indeces; Vector3[] verts; Vector2[] uv;
            subMeshes = new Mesh[geoCount];    
            for (int i = 0; i < geoCount; i++)
            {
                GetMesh(sr, out indeces, out verts, out uv);
                Mesh subMesh = new Mesh();
                subMesh.vertices = verts;
                subMesh.indeces = indeces;
                subMesh.uv = uv;
                subMeshes[i] = subMesh;
            }*/
            sr.Dispose();
        }
        void GetMesh(BinaryReader br, out int[] indeces, out Vector3[] verts, out Vector2[] uv)
        {
            indeces = new int[0];
            verts = new Vector3[0];
            uv = new Vector2[0];
            int indecesCount = br.ReadInt32();
            int vertices = br.ReadInt32();
            indeces = new int[indecesCount];
            verts = new Vector3[vertices];
            uv = new Vector2[vertices];
            for (int n = 0; n < indecesCount; n++)
            {
                indeces[n] = br.ReadInt32();
            }
            for (int n = 0; n < vertices; n++)
            {
                float x, y, z;
                x = br.ReadSingle();
                y = br.ReadSingle();
                z = -br.ReadSingle();
                verts[n] = new Vector3(x, y, z);
            }
            for (int n = 0; n < vertices; n++)
            {
                float x, y;
                x = br.ReadSingle();
                y = br.ReadSingle();
                uv[n] = new Vector2(x, y);
            }
        }
        public void CreateCube(Vector3 size)
        {
            vertices = new Vector3[8];
                                                                //FLD
            vertices[1] = new Vector3(0,size.y,0);              //FLU
            vertices[2] = new Vector3(size.x, size.y, 0);       //FRU
            vertices[3] = new Vector3(size.x, 0, 0);            //FRD
            vertices[4] = new Vector3(0,0,size.z);              //BLD
            vertices[5] = new Vector3(0, size.y, size.z);            //BLU
            vertices[6] = new Vector3(size.x, size.y, size.z);  //BRU
            vertices[7] = new Vector3(size.x,0,size.z);         //BRD

            indeces = new int[36];
            //FRONT
            indeces[0] = 0;
            indeces[1] = 1;
            indeces[2] = 2;

            indeces[3] = 0;
            indeces[4] = 2;
            indeces[5] = 3;

            //LEFT
            indeces[6] = 4;
            indeces[7] = 5;
            indeces[8] = 1;

            indeces[9] = 4;
            indeces[10] = 1;
            indeces[11] = 0;

            //RIGHT
            indeces[12] = 3;
            indeces[13] = 2;
            indeces[14] = 6;

            indeces[15] = 3;
            indeces[16] = 6;
            indeces[17] = 7;

            //BEHIND
            indeces[18] = 7;
            indeces[19] = 6;
            indeces[20] = 5;

            indeces[21] = 7;
            indeces[22] = 5;
            indeces[23] = 4;

            //UPWARD
            indeces[24] = 1;
            indeces[25] = 5;
            indeces[26] = 6;

            indeces[27] = 1;
            indeces[28] = 6;
            indeces[29] = 2;

            //DOWNWARD
            indeces[30] = 4;
            indeces[31] = 0;
            indeces[32] = 3;

            indeces[33] = 4;
            indeces[34] = 3;
            indeces[35] = 7;






        }
    }
}