﻿using System;
using System.Collections.Generic;
using System.Text;
using System.DrawingCore;
using System.Threading.Tasks;
using System.DrawingCore.Imaging;
using Matrix4x4 = System.Numerics.Matrix4x4;
using Quaternion = System.Numerics.Quaternion;


namespace Raytracer
{
    public class Renderer
    {
        Transform scene;
        int threadNum;
        public Vector2i areaSize = new Vector2i(32, 32);
        Renderable[] renderables;

        public Renderer(Transform scene, int thread)
        {
            this.scene = scene;
            threadNum = thread;
        }
        void RenderTask(Camera cam, ref RenderProperty property, System.Numerics.Vector3[,] pixels, Vector4i area)
        {
            HitInfo hf;
            int ax, ay;
            Random rnd = new Random();

            for (int x = 0; x < area.z; x++)
            {
                ax = area.x + x;
                for (int y = 0; y < area.w; y++)
                {
                    ay = area.y + y;
                    Ray ray = new Ray(cam.transform.position, new Vector3(cam.frameSize.x * ((double)ax / (double)cam.resolution.x - 0.5), -cam.frameSize.y * ((double)ay / (double)cam.resolution.y - 0.5),cam.focalDistance));
                    //Console.WriteLine(ray.direction);
                    for (int i = 0; i < property.SamplePerPixel; i++)
                    {
                        if (Trace(ray, out hf, property.IndentDepth, rnd))
                        {
                            //rnd.NextDouble();
                            pixels[ax, ay] += hf.color;
                        }
                    }
                }
            }
            Console.WriteLine("Area " + area.x + "," + area.y + " done!");
        }
        void BuildRenderables()
        {
            Renderable r;
            List<Renderable> rs = new List<Renderable>();
            foreach (var i in scene.childen)
            {
                r = i.sceneObject as Renderable;
                if (r == null) continue;
                var t = r as MeshRenderer;
                if (t != null)
                {
                    Matrix4x4 m = Matrix4x4.CreateScale(r.transform.scale);
                    m = m * Matrix4x4.CreateFromQuaternion(r.transform.rotation);
                    m = m * Matrix4x4.CreateTranslation(r.transform.position);
                    for (int n = 0; n < t.mesh.vertices.Length; n++)
                    {
                        t.mesh.vertices[n] = t.mesh.vertices[n] * m;
                    }
                }
                rs.Add(r);
            }
            renderables = rs.ToArray();
        }
        public Bitmap Render(Camera cam, RenderProperty property)
        {
            BuildRenderables();
            Bitmap bmp = new Bitmap(cam.resolution.x, cam.resolution.y, PixelFormat.Format24bppRgb);
            System.Numerics.Vector3[,] pixels = new System.Numerics.Vector3[cam.resolution.x, cam.resolution.y];
            List<Task> tasks = new List<Task>();
            int bx = cam.resolution.x / areaSize.x, by = cam.resolution.y / areaSize.y;
            int done = 0, started = 0, total = bx * by;
            do
            {
                if (started < total)
                {
                    int x, y;
                    x = started / bx;
                    y = started % bx;
                    Console.WriteLine("Area " + bx*x + "," + by*y + " started!");
                    tasks.Add(Task.Run(() =>
                    {
                        RenderTask(cam, ref property, pixels, new Vector4i(x * areaSize.x, y * areaSize.y, areaSize.x, areaSize.y));
                    }));
                }
                started++;
                if (tasks.Count >= threadNum || started + threadNum >= total)
                {
                    int i = Task.WaitAny(tasks.ToArray());
                    tasks.RemoveAt(i);
                    done++;
                }
            }
            while (done < total);

            for (int x = 0; x < cam.resolution.x; x++)
            {
                for (int y = 0; y < cam.resolution.y; y++)
                {
                    float r = pixels[x, y].X / property.SamplePerPixel;
                    float g = pixels[x, y].Y / property.SamplePerPixel;
                    float b = pixels[x, y].Z / property.SamplePerPixel;
                    if (double.IsNaN(r) || double.IsNaN(g) || double.IsNaN(b))
                    {
                        Console.WriteLine(pixels[x, y]);
                        continue;
                    }
                    bmp.SetPixel(x, y, Color.FromArgb((int)(r * 255), (int)(g * 255), (int)(b * 255)));
                }
            }
            return bmp;
        }
        bool Trace(Ray ray, out HitInfo hitinfo, int indent, Random rnd)
        {
            indent--;
            if (indent < 0)
            {
                hitinfo = new HitInfo();
                return false;
            }
            //hitinfo.color = new Vector3(1);
            Renderable r;
            HitInfo selected = new HitInfo(), nHit;
            Renderable rs = null;
            bool ret = false;
            double minDist = 999999999999;
            foreach (var i in renderables)
            {
                if (IntersectAABB(i.bound, ray))
                {
                    if (i.HitTest(ray, out hitinfo))
                    {
                        if ((hitinfo.position - ray.origin).sqrMagnitude < minDist)
                        {
                            selected = hitinfo;
                            minDist = (hitinfo.position - ray.origin).sqrMagnitude;
                            rs = i;
                            ret = true;

                        }
                    }
                }

            }
            hitinfo = selected;

            if (ret && rs != null)
            {
                Vector3 exitRay,c,illum;
                //rs.shader.rnd = rnd;
                c = rs.shader.Shade(rnd,ray.direction,hitinfo.normal,out exitRay,out illum);
                if (c.sqrMagnitude >= 0.001)
                {
                    Trace(new Ray(hitinfo.position, exitRay), out nHit, indent, rnd);
                    hitinfo.color = nHit.color * c + illum;
                }
                else
                {
                    hitinfo.color = illum;
                }
            }
            else
            {
                hitinfo.color = new Vector3(0);
            }
            return ret;
        }
        bool IntersectAABB(Bound aabb, Ray ray)
        {
            return true;
        }
    }
    public struct RenderProperty
    {
        public int SamplePerPixel;
        public int IndentDepth;
        public RenderProperty(int spp, int indd)
        {
            SamplePerPixel = spp;
            IndentDepth = indd;
        }
    }
}
